"""user facing functions for normal use"""

from __future__ import annotations
import asyncio
from pathlib import Path
from typing import Any, Callable

from luxos.asyncops import rexec  # noqa: F401
import luxos.misc


def load_ips_from_csv(path: Path | str, port: int = 4028) -> list[tuple[str, int]]:
    from luxos.scripts.luxos import load_ip_list_from_csv

    result = []
    for ip in load_ip_list_from_csv(str(path)):
        result.append((ip, port))
    return result


async def launch(
    addresses: list[tuple[str, int]], call: Callable[[str, int], Any], *args, **kwargs
) -> Any:
    """launch an async function on a list of (host, port) miners

    Eg.
        async printme(host, port, value):
            print(await rexec(host, port, "version"))
        asyncio.run(launch([("127.0.0.1", 4028)], printme, value=11, batch=10))
    """
    # special kwargs!!
    n = int(kwargs.pop("batch") or 0) if "batch" in kwargs else None
    if n and n < 0:
        raise RuntimeError(
            f"cannot pass the 'batch' keyword argument with a value < 0: batch={n}"
        )

    if n:
        result = []
        for subaddresses in luxos.misc.batched(addresses, n):
            tasks = [call(*address, *args, **kwargs) for address in subaddresses]
            for task in await asyncio.gather(*tasks, return_exceptions=True):
                result.append(task)
        return result
    else:
        tasks = [call(*address, *args, **kwargs) for address in addresses]
        return await asyncio.gather(*tasks, return_exceptions=True)



from __future__ import annotations
import asyncio
import json

from .. import misc
from .. import asyncops


async def run(
    ipaddresses: list[str],
    port: int,
    cmd: str,
    params: list[str],
    timeout: float,
    delay: float | None,
    details: bool,
    batchsize: int = 0,
) -> None:
    result = {}
    batchsize = max(batchsize, 2)

    for grupid, addresses in enumerate(
        misc.batched([(ip, port) for ip in ipaddresses], n=batchsize)
    ):
        tasks = []
        for host, port in addresses:
            tasks.append(
                asyncops.execute_command(
                    host, port, timeout, cmd, params, add_address=True
                )
            )
        result[grupid] = await asyncio.gather(*tasks, return_exceptions=True)

        # runs only on batchsize, wait delay then proceed onto the next batch
        if delay:
            await asyncio.sleep(delay)

    alltasks = [task for group in result.values() for task in group]
    successes = [task for task in alltasks if not isinstance(task, Exception)]
    failures = [task for task in alltasks if isinstance(task, Exception)]

    # print a nice report
    print(f"task executed sucessfully (use -a|--all for details): {len(successes)}")
    if details:
        for (host, port), task in successes:  # type: ignore
            print(f"  > {host}:{port}")
            print(misc.indent(json.dumps(task, indent=2, sort_keys=True), pre="  | "))
    print(f"task executed failures: {len(failures)}")
    for failure in failures:
        print(f"  {failure}")


def main(*args, **kwargs) -> None:
    asyncio.run(run(*args, **kwargs))



import argparse
import sys
from unittest import mock

from luxos.cli import v1 as cli


def test_docstring():
    @cli.cli()
    def hello():
        "this is a docstring"
        pass

    assert hello.__doc__ == "this is a docstring"


def test_callme():
    @cli.cli()
    def main(parser):
        assert isinstance(parser, argparse.ArgumentParser)

    main()

    args = ["dummy.py"]
    with mock.patch.object(sys, "argv", args):

        @cli.cli()
        def main2(args):
            assert isinstance(args, argparse.Namespace)

        main2()



from __future__ import annotations

import os
import asyncio
import pytest
import luxos.asyncops as aapi
from luxos import exceptions
from luxos import misc

## NOTE ##
# This tests spawn an underlying server, it might be better not run
# unattended. Some also require a miner, we might not have it handy.
pytestmark = pytest.mark.manual


def getminer() -> None | tuple[str, int]:
    if not (minerd := os.getenv("LUXOS_TEST_MINER")):
        return None
    host, port = minerd.split(":")
    return host.strip(), int(port)


def test_rexec_paramteres():
    assert aapi._rexec_paramteres(None) == []
    assert aapi._rexec_paramteres("hello") == ["hello"]
    assert aapi._rexec_paramteres(["hello", "world"]) == ["hello", "world"]
    assert aapi._rexec_paramteres(["hello", 1]) == ["hello", "1"]
    assert aapi._rexec_paramteres({"hello": 1}) == ["hello=1"]
    assert aapi._rexec_paramteres({"hello": True}) == ["hello=true"]


def test_validate_message():
    pytest.raises(
        exceptions.MinerCommandMalformedMessageError, aapi.validate_message, "a", 0, {}
    )
    host, port = "a", 0

    assert aapi.validate_message(host, port, {"STATUS": 1, "id": 2})

    pytest.raises(
        exceptions.MinerCommandMalformedMessageError,
        aapi.validate_message,
        host,
        port,
        {
            "STATUS": 1,
            "id": 2,
        },
        "wooow",
    )

    assert aapi.validate_message(
        host, port, {"STATUS": 1, "id": 2, "wooow": [1, 2]}, "wooow"
    )

    with pytest.raises(exceptions.MinerCommandMalformedMessageError) as excinfo:
        aapi.validate_message(
            host, port, {"STATUS": 1, "id": 2, "wooow": [1, 2]}, "wooow", minfields=9
        )
    assert excinfo.value.args[2] == "found 2 fields for wooow invalid: 2 <= 9"

    with pytest.raises(exceptions.MinerCommandMalformedMessageError) as excinfo:
        aapi.validate_message(
            host, port, {"STATUS": 1, "id": 2, "wooow": [1, 2]}, "wooow", maxfields=1
        )
    assert excinfo.value.args[2] == "found 2 fields for wooow invalid: 2 >= 1"

    with pytest.raises(exceptions.MinerCommandMalformedMessageError) as excinfo:
        aapi.validate_message(
            host,
            port,
            {"STATUS": 1, "id": 2, "wooow": [1, 2]},
            "wooow",
            minfields=9,
            maxfields=10,
        )
    assert excinfo.value.args[2] == "found 2 fields for wooow invalid: 2 <= 9"

    assert aapi.validate_message(
        host,
        port,
        {"STATUS": 1, "id": 2, "wooow": [1, 2]},
        "wooow",
        minfields=2,
        maxfields=10,
    )


@pytest.mark.asyncio
async def test_private_roundtrip_one_listener(echopool):
    """checks roundrtip sends and receive a message (1-listener)"""
    echopool.start(1, mode="echo+")
    host, port = echopool.addresses[0]
    ret = await aapi._roundtrip(host, port, "hello", None)
    assert ret == f"received by ('{host}', {port}): hello"


@pytest.mark.asyncio
async def test_private_roundtrip_many_listeners(echopool):
    """checks the roundrip can connect en-masse to many lsiteners"""
    echopool.start(100, mode="echo+")

    blocks = {}
    for index, group in enumerate(misc.batched(echopool.addresses, 10)):
        tasks = []
        for host, port in group:
            tasks.append(aapi._roundtrip(host, port, "hello olleh", None))
        blocks[index] = await asyncio.gather(*tasks, return_exceptions=True)

    assert len(blocks) == 10
    assert sum(len(x) for x in blocks.values()) == 100
    allitems = [item for b in blocks.values() for item in b]
    assert f"received by {echopool.addresses[3]}: hello olleh" in allitems


@pytest.mark.skipif(not getminer(), reason="need to set LUXOS_TEST_MINER")
@pytest.mark.asyncio
async def test_miner_logon_logoff_cycle():
    host, port = getminer()

    sid = None
    try:
        sid = await aapi.logon(host, port)
    except exceptions.MinerCommandSessionAlreadyActive as e:
        raise RuntimeError("a session is already active on {host}:{port}") from e
    finally:
        if sid:
            await aapi.logoff(host, port, sid)


@pytest.mark.skipif(not getminer(), reason="need to set LUXOS_TEST_MINER")
@pytest.mark.asyncio
async def test_miner_double_logon_cycle():
    host, port = getminer()

    sid = await aapi.logon(host, port)
    try:
        with pytest.raises(exceptions.MinerCommandSessionAlreadyActive) as excinfo:
            await aapi.logon(host, port)
        assert excinfo.value.args[0] == f"session active for {host}:{port}"
    except Exception:
        pass
    finally:
        if sid:
            await aapi.logoff(host, port, sid)


@pytest.mark.skipif(not getminer(), reason="need to set LUXOS_TEST_MINER")
@pytest.mark.asyncio
async def test_miner_version():
    host, port = getminer()

    res = await aapi.execute_command(host, port, None, "version", asjson=True)
    assert "VERSION" in res
    assert len(res["VERSION"]) == 1
    assert "API" in res["VERSION"][0]


@pytest.mark.skipif(not getminer(), reason="need to set LUXOS_TEST_MINER")
@pytest.mark.asyncio
async def test_miner_profile_sets():
    from string import ascii_lowercase
    from random import choices

    # random profile name
    profile = f"test-{''.join(choices(ascii_lowercase, k=5))}"

    # get the initial profile list
    host, port = getminer()
    profiles = (await aapi.rexec(host, port, "profiles"))["PROFILES"]
    assert profiles and profile not in {p["Profile Name"] for p in profiles}

    # create a new profile
    # TODO there's a bug
    #  when ATM is running you shouldn't be able to create/delete
    #  profiles, the following test shoudl fail at the first try
    # -> to pass this test, you need to disable ATM

    params = f"{profile},{profiles[0]['Frequency']},{profiles[0]['Voltage']}"
    await aapi.rexec(host, port, "profilenew", params)

    try:
        profiles1 = (await aapi.rexec(host, port, "profiles"))["PROFILES"]
        assert profile in {p["Profile Name"] for p in profiles1}
    finally:
        await aapi.rexec(host, port, "profilerem", profile)

    # verify we restored the same profiles
    profiles2 = (await aapi.rexec(host, port, "profiles"))["PROFILES"]
    expected = {p["Profile Name"] for p in profiles}
    found = {p["Profile Name"] for p in profiles2}
    assert found == expected



from __future__ import annotations

import logging
import json
import importlib.resources

log = logging.getLogger(__name__)


COMMANDS = json.loads(
    (importlib.resources.files("luxos") / "api.json")
    .read_text()
)


def logon_required(cmd: str, commands_list=COMMANDS) -> bool | None:
    # Check if the command requires logon to LuxOS API

    if cmd not in COMMANDS:
        log.info("%s command is not supported. "
                 "Try again with a different command.", cmd)
        return None

    return COMMANDS[cmd]["logon_required"]


# TODO timeouts should be float | None
def execute_command(host: str, port: int, timeout: int, cmd: str, params: list[str], verbose: bool):
    from .scripts import luxos
    return luxos.execute_command(host, port, timeout, cmd, params, verbose)


